#!/bin/bash
#
# edited by jK (7.7.05)
#
# based upon "MythStreamTV Install Script"
InstallFile() {
   SOURCE_FILE="$1"
   DEST_FILE="$2"

   echo -n ".... Installing $DEST_FILE: "
   cat $SOURCE_FILE                                             | \
         sed -e "s|#\*\*NEW_STREAM_DIR\*\*#|$NEW_STREAM_DIR|g"  | \
         sed -e "s|#\*\*STREAM_ADDR\*\*#|$STREAM_ADDR|g"        | \
         sed -e "s|#\*\*STREAM_ADMIN\*\*#|$STREAM_ADMIN|g"      | \
         sed -e "s|#\*\*STREAM_PORT\*\*#|$STREAM_PORT|g"        | \
         sed -e "s|#\*\*VLC_PATH\*\*#|$VLC_PATH|g"              > $DEST_FILE
   RET=$?
   if [ $RET -gt 0 ]; then
      echo Failed
      exit 1
   else
      echo Success
   fi
}

RunAndTest() {
   $*
   RET=$?
   if [ $RET -gt 0 ]; then
      echo Failed
      exit 1
   fi
}

GetString() {
   echo -n "$1? [$2] "
   read
   if [ -n "$REPLY" ]; then
      READ_STRING="$REPLY"
   else
      READ_STRING="$2"
   fi
}

GetYN() {
   GetString "$1" "$2"
   case $READ_STRING in
      y | Y | yes | YES)
         return 1
         ;;
   esac

   return 0;
}

InstallToNewMythWeb() {
   echo -n "........ Fixing up themes/default/theme.php: "
   OUT=`cat $HTTP_DIR/themes/Default/theme.php | \
                                           grep -n "MythVideo"`
   RET=$?
   if [ $RET -gt 0 ]; then
      echo "Internal error (grep)"
      exit 1
   fi
   let num="`echo $OUT | cut -d ':' -f 1`"
   let end="$num - 1"
   cat $HTTP_DIR/themes/Default/theme.php | head -n $end \
                                      > $HTTP_DIR/themes/Default/theme.php.new
   RET=$?
   if [ $RET -gt 0 ]; then
      echo Failed
      exit 1
   fi
   cat << EOF >> $HTTP_DIR/themes/Default/theme.php.new
        <a id="stream_link"<?php if (section == 'stream') echo ' class="current_section"' ?> href="mythstream.php" onmouseover="return help_text('<?php echo str_replace("'", "\\'", t('MythStream on the web.')) ?>')" onmouseout="return help_text()">
            <img src="<?php echo theme_dir ?>img/radio.png" width="48" height="48" class="alpha_png" alt="MythStream" />
        </a>
EOF
   cat $HTTP_DIR/themes/Default/theme.php | tail -n +$num \
                                      >> $HTTP_DIR/themes/Default/theme.php.new
   RET=$?
   if [ $RET -gt 0 ]; then
      echo Failed
      exit 1
   fi
   RunAndTest cp -f $HTTP_DIR/themes/Default/theme.php.new \
                    $HTTP_DIR/themes/Default/theme.php
   rm -f $HTTP_DIR/themes/Default/theme.php.new
   echo Done



   echo -n "........ Fixing up languages/English.php: "
   OUT=`cat $HTTP_DIR/languages/English.php | grep -n "MythVideo"`
   RET=$?
   if [ $RET -gt 0 ]; then
      echo "Internal error (grep)"
      exit 1
   fi
   let num="`echo $OUT | cut -d ':' -f 1`"
   let begin="$num - 1"
   cat $HTTP_DIR/languages/English.php | head -n $begin  \
                                      > $HTTP_DIR/languages/English.php.new
   RET=$?
   if [ $RET -gt 0 ]; then
      echo Failed
      exit 1
   fi
   cat << EOF >> $HTTP_DIR/languages/English.php.new
    'MythStream on the web.'         => '',
EOF
   cat $HTTP_DIR/languages/English.php | tail -n $num \
                                      >> $HTTP_DIR/languages/English.php.new
   RET=$?
   if [ $RET -gt 0 ]; then
      echo Failed
      exit 1
   fi
   RunAndTest cp -f $HTTP_DIR/languages/English.php.new \
                    $HTTP_DIR/languages/English.php
   rm -f $HTTP_DIR/languages/English.php.new
   echo Done
}



echo "Starting Configuration....."

# Set our defaults
HERE_DIR=`pwd`
HTTP_DIR=/var/www/localhost/htdocs/mythweb

# Ask for configuration
let not_ok=1
while [ $not_ok -gt 0 ]; do
   echo
   GetString "MythWeb HTTP Directory" "$HTTP_DIR"
   HTTP_DIR="$READ_STRING"

   echo
   echo "Please confirm your configuration:"
   echo "----------------------------------------------------------------------"
   echo "    MythWeb HTTP Directory         : $HTTP_DIR"
   echo "----------------------------------------------------------------------"
   GetYN "Is this correct" "n"
   RET=$?
   if [ $RET -eq 1 ]; then
      not_ok=0
   fi
done



echo
echo "Installing MythStream WebInterface........."
   
echo -n ".... Checking MythWeb Compatablity: "
# Let's see if it is the new or old one
cat $HTTP_DIR/themes/Default/theme.php | \
                                  grep -n "_LANG_MANUALLY_SCHEDULE" > /dev/null
RET=$?
if [ $RET -gt 0 ]; then
   cat $HTTP_DIR/themes/Default/theme.php | \
                                  grep -n "Recorded Programs" > /dev/null
   RET=$?
   if [ $RET -gt 0 ]; then
      echo "Oops! No support for this version of MythWeb"
      exit 1
   fi
   NEW_MYTHWEB=1
   echo "Good"
else
   NEW_MYTHWEB=0
   echo "Oops! No support for this version of MythWeb"
   exit 1
fi

InstallFile $HERE_DIR/stream.php \
            $HTTP_DIR/mythstream.php

InstallFile $HERE_DIR/stream_edit.php \
            $HTTP_DIR/stream_edit.php

InstallFile $HERE_DIR/mythstream.php \
            $HTTP_DIR/themes/Default/mythstream.php

InstallFile $HERE_DIR/mythstream_editmode.php \
            $HTTP_DIR/themes/Default/mythstream_editmode.php

InstallFile $HERE_DIR/radio6.png \
            $HTTP_DIR/themes/Default/img/radio.png

echo -n ".... Checking MythWeb Installation: "
cat $HTTP_DIR/themes/Default/theme.php | grep "mythstream.php" > /dev/null
RET=$?
if [ $RET -gt 0 ]; then
   echo "Required - Installing..."
   if [ $NEW_MYTHWEB -eq 1 ]; then
      InstallToNewMythWeb
   fi
else
   echo "Not required, skipping"
fi

echo -n ".... Fixing up file permissions: "
RunAndTest chown -R root:root $HTTP_DIR/mythstream*.php
RunAndTest chown -R root:root $HTTP_DIR/themes/Default/mythstream*.php
echo Done

echo
echo "Installation Complete!  Good Luck."
